--[[
--
-- TODO
-- * Add a countdown to when the next person in line should get close
--
--]]

local f = CreateFrame("Frame")

------------------------------------------------------------------------
-- Locals
--

local whisperNext = "TMP: You are getting the Spark next after %s!"
local whisperNow = "TMP: GET IT NOW!"
local warningNotScanned = "%s could not be scanned, you need to scan again with /tmpscan when that player is closer."
local warningMoreToScan = "Note that the whole raid hasn't been scanned for roles yet. This will happen automatically over time in most cases, or you can force it with /tmpscan."
local warningScanDone = "TMP scan has completed."
local warningScanStarting = "TormentMePlz role scan in progress, please wait ..."

-- Set when we zone in to either 10 or 25
local raidType = nil

-- Needs to be file-local because we use it to check whether or not
-- all the raid members have been scanned.
local needsScanning = setmetatable({}, {__newindex = function(self, key, value)
	for i, v in next, self do
		-- Make sure we don't try to scan the same person twice.
		if v == value then return end
	end
	rawset(self, key, value)
end})

-- key: player, value: assigment
-- assignment: Ranged = 2, Melee = 1, Tank = 3, Healer = 4
--[[local assignments = setmetatable({}, {__newindex = function(t, key, value)
	print(key .. " assigned as " .. tostring(value))
	rawset(t, key, value)
end})]]
local assignments = {}

------------------------------------------------------------------------
-- Proximity scanning
--

do
	local roomHeight, roomWidth = 1587.49993896484, 1058.3332824707
	local function getUnitPosition(unit)
		local x,y = GetPlayerMapPosition(unit)
		if x <= 0 and y <= 0 then
			SetMapToCurrentZone()
			x,y = GetPlayerMapPosition(unit)
		end
		return x,y
	end
	local function getDistance(unit1,unit2)
		local x1,y1 = getUnitPosition(unit1)
		local x2,y2 = getUnitPosition(unit2)
		local dx = (x2 - x1) * roomWidth
		local dy = (y2 - y1) * roomHeight
		return (dx*dx + dy*dy)^(0.5) -- dx*dx is faster than dx^2
	end

	local tormentedPlayers = {}
	local nextTargets = {}

	local tmpDistances = {}
	local possibleTargets = {}
	local function distanceSort(a, b) return (tmpDistances[a] or 0) < (tmpDistances[b] or 0) end
	local function durationSort(a, b) return (tormentedPlayers[a] or 0) < (tormentedPlayers[b] or 0) end
	local function isAcceptable(first, second)
		if raidType == 10 then return true end
		return first == second
	end
	local function getNextTarget(current, assignment)
		if assignment == 3 then assignment = 1 end -- Tank -> Melee
		if assignment == 4 then assignment = 2 end -- Healer -> Ranged
		wipe(tmpDistances)
		wipe(possibleTargets)
		for player, a in pairs(assignments) do
			if current ~= player and isAcceptable(a, assignment) and not UnitIsDeadOrGhost(player) and not tormentedPlayers[player] then
				tmpDistances[player] = getDistance(current, player)
				possibleTargets[#possibleTargets + 1] = player
			end
		end
		table.sort(possibleTargets, distanceSort)
		if #possibleTargets == 0 then
			print("All possible targets are debuffed, returning the one with least duration.")
			-- There are no possible targets, let's find the player with the shortest time
			-- left until his Tormented debuff expires.
			for player, a in pairs(assignments) do
				if current ~= player and isAcceptable(a, assignment) and not UnitIsDeadOrGhost(player) then
					possibleTargets[#possibleTargets + 1] = player
				end
			end
			table.sort(possibleTargets, durationSort)
		end
		return possibleTargets[1]
	end

	local function populateNextTarget(player)
		local nextTarget = getNextTarget(player, assignments[player])
		if nextTarget then
			print(tostring(nextTarget) .. " is next after " .. tostring(player) .. ".")
			nextTargets[player] = nextTarget
			SendChatMessage(whisperNext:format(player), "WHISPER", nil, nextTarget)
		else
			print("Could not find a replacement for " .. tostring(player) .. ".")
		end
	end

	local tormentedIds = {
		[99489] = true,
		[99257] = true,
		[99402] = true,
		[99403] = true,
		[99404] = true,
	}

	local tormentIds = {
		[99256] = true,
		[100230] = true,
		[100231] = true,
		[100232] = true,
	}
	local durations = {[10] = 40, [25] = 60}

-- SPELL_AURA_APPLIED,0xF130D0F70000C84E,"Shard of Torment",0xa48,0x0,0x0400000001F3B9D3,"XXX",0x514,0x0,100230,"Torment",0x20,DEBUFF
-- SPELL_AURA_APPLIED_DOSE,0xF130D0F70000C84E,"Shard of Torment",0xa48,0x0,0x0400000001F3B9D3,"XXX",0x514,0x0,100230,"Torment",0x20,DEBUFF,2
-- SPELL_AURA_REMOVED,0xF130D0F70000C84F,"Shard of Torment",0xa48,0x0,0x0400000003D66AF7,"XXX",0x512,0x0,100230,"Torment",0x20,DEBUFF
-- SPELL_AURA_APPLIED,0x0400000003D66AF7,"XXX",0x512,0x0,0x0400000003D66AF7,"XXX",0x512,0x0,99402,"Tormented",0x1,DEBUFF
-- SPELL_AURA_REMOVED,0x0400000003D66AF7,"XXX",0x512,0x0,0x0400000003D66AF7,"XXX",0x512,0x0,99402,"Tormented",0x1,DEBUFF

	function f:COMBAT_LOG_EVENT_UNFILTERED(_, _, event, _, _, _, _, _, _, player, _, _, spellId, _, _, _, buffStack)
		if event == "SPELL_AURA_APPLIED" then
			if tormentIds[spellId] then
				for k, v in pairs(nextTargets) do
					if v == player then
						nextTargets[k] = nil
					end
				end
			elseif tormentedIds[spellId] then
				tormentedPlayers[player] = GetTime() + durations[raidType]
			end
		elseif event == "SPELL_AURA_APPLIED_DOSE" then
			if not tormentIds[spellId] then return end
			if not nextTargets[player] then
				populateNextTarget(player)
			end
			if buffStack > 7 then
				if not nextTargets[player] then
					print(player .. " has more than 7 stacks but there is noone assigned to replace him?!")
					return
				end
				print("Telling " .. nextTargets[player] .. " to take over for " .. player .. " NOW.")
				SendChatMessage(whisperNow, "WHISPER", nil, nextTargets[player])
			end
		elseif event == "SPELL_AURA_REMOVED" then
			if not tormentedIds[spellId] then return end
			tormentedPlayers[player] = nil
		end
	end
end

------------------------------------------------------------------------
-- Assignment and inspect handling
--

do
	local lastUnitQueried = nil
	local lastScanTime = 0
	local total = 0

	hooksecurefunc("NotifyInspect", function(unit)
		lastUnitQueried = UnitName(unit)
		lastScanTime = GetTime()
	end)

	f:SetScript("OnShow", function() print(warningScanStarting) end)
	f:SetScript("OnUpdate", function(self, elapsed)
		total = total + elapsed
		if total >= 4 then
			total = 0
			if lastUnitQueried then return end
			local unit = table.remove(needsScanning, 1)
			if not unit then
				print(warningScanDone)
				self:Hide()
				return
			end
			if not UnitExists(unit) then return end -- Give up on this unit entirely
			if GetTime() > (lastScanTime + 3)
				and UnitIsVisible(unit)
				and UnitIsConnected(unit)
				and CheckInteractDistance(unit, 4)
				and CanInspect(unit) then
				NotifyInspect(unit)
			else
				-- Add the unit to the end of the list so we try it again later.
				needsScanning[#needsScanning + 1] = unit
			end
		end
	end)
	f:Hide()

	local function getUnitNameFromGUID(guid)
		if UnitInRaid("player") then
			for i = 1, GetNumRaidMembers() do
				local n = GetRaidRosterInfo(i)
				if n and UnitGUID(n) == guid then return n end
			end
		else
			for i = 1, GetNumPartyMembers() do
				local pId = UnitGUID("party" .. i)
				if pId == guid then return UnitName("party" .. i) end
			end
		end
	end

	-- Ranged = 2, Melee = 1, Tank = 3, Healer = 4
	local defaultAssignments = {
		MAGE = 2,
		WARLOCK = 2,
		HUNTER = 2,
		ROGUE = 1,
		WARRIOR = function(player, talentSpec, role)
			if role == "DAMAGER" or talentSpec == 1 or talentSpec == 2 then return 1 end
			if role == "TANK" or talentSpec == 3 then return 3 end
			return 1
		end,
		PALADIN = function(player, talentSpec, role)
			if role == "DAMAGER" or talentSpec == 3 then return 1 end
			if role == "TANK" or talentSpec == 2 then return 3 end
			if role == "HEALER" or talentSpec == 1 then return 4 end
			return 4
		end,
		DEATHKNIGHT = function(player, talentSpec, role)
			if role == "DAMAGER" or talentSpec == 2 or talentSpec == 3 then return 1 end
			if role == "TANK" or talentSpec == 1 then return 3 end
			return 1
		end,
		DRUID = function(player, talentSpec, role)
			if role == "DAMAGER" then
				if talentSpec == 1 then return 2
				elseif talentSpec == 2 then return 1
				end
			elseif role == "TANK" or talentSpec == 2 then
				return 3 -- Assume Feral spec means tanking.
			elseif role == "HEALER" or talentSpec == 3 then
				return 4
			elseif talentSpec == 1 then
				return 2
			end
		end,
		SHAMAN = function(player, talentSpec, role)
			if talentSpec == 1 then return 2 end
			if talentSpec == 2 then return 1 end
			if talentSpec == 3 or role == "HEALER" then return 4 end
			if role == "DAMAGER" then return 2 end -- Assume shamans are elemental
		end,
		PRIEST = function(player, talentSpec, role)
			if role == "DAMAGER" or talentSpec == 3 then return 2 end
			return 4
		end,
	}
	local talentSpecs = {}

	local function assignUnit(name, class, override)
		if not override then override = UnitGroupRolesAssigned(name) end
		local def = defaultAssignments[class]
		if type(def) == "function" then
			local a = def(name, talentSpecs[name], override)
			if a then
				assignments[name] = a
			elseif lastUnitQueried ~= name then
				needsScanning[#needsScanning + 1] = name
				assignments[name] = nil
			end
		else
			assignments[name] = def
		end
	end

	function f:INSPECT_READY(event, guid)
		if not lastUnitQueried or guid ~= UnitGUID(lastUnitQueried) then
			lastUnitQueried = getUnitNameFromGUID(guid)
		end
		if not lastUnitQueried then return end
		local group = GetActiveTalentGroup(true)
		local tree = talentSpecs[lastUnitQueried]
		for i = 1, 3 do
			local _,_,_,_, pointsSpent = GetTalentTabInfo(i, true, nil, group)
			if pointsSpent > 30 then
				tree = i
				break
			end
		end
		talentSpecs[lastUnitQueried] = tree
		local _, class = UnitClass(lastUnitQueried)
		if class then
			assignUnit(lastUnitQueried, class)
			for i, v in next, needsScanning do
				if v == lastUnitQueried then
					table.remove(needsScanning, i)
					break
				end
			end
		end
		lastUnitQueried = nil
	end

	do
		local specChange = {}
		for specIndex, spellId in next, _G.TALENT_ACTIVATION_SPELLS do
			specChange[GetSpellInfo(spellId)] = specIndex
		end
		function f:UNIT_SPELLCAST_SUCCEEDED(event, unit, spell)
			if not specChange[spell] or not UnitInRaid(unit) then return end
			local _, class = UnitClass(unit)
			local name = UnitName(unit)
			talentSpecs[name] = nil
			assignUnit(name, class)
		end
	end

	function f:ROLE_CHANGED_INFORM(event, target, source, oldRole, newRole)
		local _, class = UnitClass(target)
		-- XXX We need to pass the new role in because
		-- UnitGroupRolesAssigned still returns the old value when this
		-- event is triggered ...
		assignUnit(target, class, newRole)
	end

	function f:RAID_ROSTER_UPDATE()
		for i = 1, GetNumRaidMembers() do
			local n,_,_,_,_,class = GetRaidRosterInfo(i)
			if n and not assignments[n] then
				assignUnit(n, class)
			end
		end
		for u in pairs(assignments) do
			if not UnitInRaid(u) then
				assignments[u] = nil
				talentSpecs[u] = nil
			end
		end
	end
end

------------------------------------------------------------------------
-- Warn user if we're not ready
--

do
	local function checkBaleroc(unit)
		local guid = UnitGUID(unit)
		if not guid then return end
		local id = tonumber(guid:sub(7, 10), 16)
		if id ~= 53494 then return end -- We're not engaging Baleroc
		if #needsScanning > 0 then
			print(warningMoreToScan)
		end
	end
	function f:INSTANCE_ENCOUNTER_ENGAGE_UNIT() checkBaleroc("boss1") end
	function f:PLAYER_TARGET_CHANGED() checkBaleroc("target") end
end

------------------------------------------------------------------------
-- Addon setup
--

function f:ZONE_CHANGED_NEW_AREA()
	local id = GetCurrentMapAreaID()
	if id == 800 then -- 800 = Firelands
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		local _,_, diff = GetInstanceInfo()
		if diff == 1 or diff == 3 then
			raidType = 10
		elseif diff == 2 or diff == 4 then
			raidType = 25
		end
		print("TMP enabled for " .. tostring(raidType) .. "man.")
	else
		raidType = nil
		self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		self:UnregisterEvent("PLAYER_TARGET_CHANGED")
		self:UnregisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end

f:RegisterEvent("ROLE_CHANGED_INFORM")
f:RegisterEvent("ZONE_CHANGED_NEW_AREA")
f:RegisterEvent("INSPECT_READY")
f:RegisterEvent("RAID_ROSTER_UPDATE")
f:SetScript("OnEvent", function(self, event, ...) self[event](self, event, ...) end)

local function filter(self, event, msg)
	if msg:find("^TMP: ") then return true end
end
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", filter)

SlashCmdList.TormentMePlzScan = function() f:Show() end
SLASH_TormentMePlzScan1 = "/tmpscan"

SlashCmdList.TormentMePlz = function(input)
	local command, argument = input:match("^(%w+)%s(%w+)$")
	if not command then print("/tmp [ranged|melee|reset] <player name>"); return end
	if not UnitExists(argument) then print("Invalid unit " .. tostring(argument)); return end
	if command == "ranged" then
		assignments[argument] = 2
	elseif command == "melee" then
		assignments[argument] = 1
	else
		assignments[argument] = nil
	end
	print("Set " .. tostring(argument) .. " to " .. tostring(command) .. ".")
end
SLASH_TormentMePlz1 = "/tmp"

------------------------------------------------------------------------
-- LDB display
--

local ldb = LibStub("LibDataBroker-1.1", true)
if ldb then
local countFormat = "(#%d)"
local tmpMelee = {}
local tmpRanged = {}
local tmpOthers = {}
local function updateTooltip(tt)
	tt:AddLine("TormentMePlz")
	if #needsScanning > 0 then
		tt:AddDoubleLine("Needs scanning", countFormat:format(#needsScanning))
		tt:AddLine(table.concat(needsScanning, ", "), 1, 1, 1, 1)
		tt:AddLine(" ")
	end
	wipe(tmpMelee); wipe(tmpRanged); wipe(tmpOthers)
	for player, a in pairs(assignments) do
		if a == 1 then
			tmpMelee[#tmpMelee+1] = player
		elseif a == 2 then
			tmpRanged[#tmpRanged+1] = player
		else
			tmpOthers[#tmpOthers+1] = player
		end
	end
	if #tmpMelee > 0 then
		tt:AddDoubleLine(MELEE or "Melee", countFormat:format(#tmpMelee))
		tt:AddLine(table.concat(tmpMelee, ", "), 1, 1, 1, 1)
		tt:AddLine(" ")
	end
	if #tmpRanged > 0 then
		tt:AddDoubleLine(RANGED or "Ranged", countFormat:format(#tmpRanged))
		tt:AddLine(table.concat(tmpRanged, ", "), 1, 1, 1, 1)
		tt:AddLine(" ")
	end
	if #tmpOthers > 0 then
		tt:AddDoubleLine("Tanks/Healers", countFormat:format(#tmpOthers))
		tt:AddLine(table.concat(tmpOthers, ", "), 1, 1, 1, 1)
		tt:AddLine(" ")
	end
	tt:AddLine("Remember to make sure people know that the whispers will be incoming, and that they know what to do when they get the different whispers.", 0.2, 1, 0.2, 1)
end
local icon = select(3, GetSpellInfo(99256))
local broker = ldb:NewDataObject("TormentMePlz", {
	type = "data source",
	icon = icon,
	text = "TormentMePlz",
	OnTooltipShow = updateTooltip,
})
end
